# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from typing import Optional, List
from odxtools.utils import read_description_from_odx
from .odxlink import OdxLinkRef, OdxLinkId, OdxDocFragment, OdxLinkDatabase

@dataclass()
class Audience:
    enabled_audience_refs: list = field(default_factory=list)
    disabled_audience_refs: list = field(default_factory=list)
    is_supplier: bool = True
    is_development: bool = True
    is_manufacturing: bool = True
    is_aftersales: bool = True
    is_aftermarket: bool = True

    _enabled_audiences: Optional[List["Audience"]] = None
    _disabled_audiences: Optional[List["Audience"]] = None

    @property
    def enabled_audiences(self):
        return self._enabled_audiences

    @property
    def disabled_audiences(self):
        return self._disabled_audiences

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        self._enabled_audiences = [odxlinks.resolve(ref)
                                   for ref in self.enabled_audience_refs]
        self._disabled_audiences = [odxlinks.resolve(ref)
                                    for ref in self.disabled_audience_refs]


@dataclass()
class AdditionalAudience:
    """
    Corresponds to ADDITIONAL-AUDIENCE.
    """
    odx_id: OdxLinkId
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None


def read_audience_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    enabled_audience_refs = [OdxLinkRef.from_et(ref, doc_frags)
        for ref in et_element.iterfind("ENABLED-AUDIENCE-REFS/"
                                       "ENABLED-AUDIENCE-REF")]
    disabled_audience_refs = [OdxLinkRef.from_et(ref, doc_frags)
        for ref in et_element.iterfind("DISABLED-AUDIENCE-REFS/"
                                       "DISABLED-AUDIENCE-REF")]
    is_supplier = et_element.get("IS-SUPPLIER", "true") == 'true'
    is_development = et_element.get("IS-DEVELOPMENT", "true") == 'true'
    is_manufacturing = et_element.get("IS-MANUFACTURING", "true") == 'true'
    is_aftersales = et_element.get("IS-AFTERSALES", "true") == 'true'
    is_aftermarket = et_element.get("IS-AFTERMARKET", "true") == 'true'

    return Audience(enabled_audience_refs=enabled_audience_refs,
                    disabled_audience_refs=disabled_audience_refs,
                    is_supplier=is_supplier,
                    is_development=is_development,
                    is_manufacturing=is_manufacturing,
                    is_aftersales=is_aftersales,
                    is_aftermarket=is_aftermarket)


def read_additional_audience_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    short_name = et_element.findtext("SHORT-NAME")
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None

    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))

    return AdditionalAudience(odx_id=odx_id,
                              short_name=short_name,
                              long_name=long_name,
                              description=description)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .database import Database
from .globals import logger

def load_odx_d_file(odx_d_file_name: str):
    container = Database(odx_d_file_name=odx_d_file_name)
    logger.info(f"--- --- --- Done with parsing --- --- ---")
    return container



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass, field
from typing import Dict, List, Any, Literal, Optional, Union

from .utils import short_name_as_id
from .nameditemlist import NamedItemList
from .utils import read_description_from_odx
from .odxlink import OdxLinkRef, OdxLinkId, OdxLinkDatabase, OdxDocFragment

UnitGroupCategory = Literal["COUNTRY", "EQUIV-UNITS"]


@dataclass
class PhysicalDimension:
    """A physical dimension is a formal definition of a unit.

    It consists of the exponents for the SI units:

    | Symbol | Name     | Quantity                  | Property                 |
    |  ---   | ---      | ---                       | ---                      |
    | s      | second   | time                      | `time_exp`               |
    | m      | metre    | length                    | `length_exp`             |
    | kg     | kilogram | mass                      | `mass_exp`               |
    | A      | ampere   | electric current          | `current_exp`            |
    | K      | kelvin   | thermodynamic temperature | `temperature_exp`        |
    | mol    | mole     | amount of substance       | `molar_amount_exp`       |
    | cd     | candela  | luminous intensity        | `luminous_intensity_exp` |

    (The first three columns are from https://en.wikipedia.org/wiki/International_System_of_Units.)


    Examples
    --------

    The unit `m/s` (or `m**1 * s**(-1)`) can be represented as
    ```
    PhysicalDimension(
        odx_id="velocity",
        short_name="metre_per_second",
        length_exp=1,
        time_exp=-1
    )
    ```
    """
    odx_id: OdxLinkId
    short_name: str
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    length_exp: int = 0
    mass_exp: int = 0
    time_exp: int = 0
    current_exp: int = 0
    temperature_exp: int = 0
    molar_amount_exp: int = 0
    luminous_intensity_exp: int = 0


@dataclass
class Unit:
    """
    A unit consists of an ID, short name and a display name.

    Additionally, a unit may reference an SI unit (`.physical_dimension`)
    and an offset to that unit (`factor_si_to_unit`, `offset_si_to_unit`).
    The factor and offset are defined such that the following equation holds true:

    UNIT = FACTOR-SI-TO-UNIT * SI-UNIT + OFFSET-SI-TO-UNIT

    For example: 1km = 1000 * 1m + 0

    Examples
    --------

    A minimal unit representing kilometres:

    ```
    Unit(
        odx_id="kilometre",
        short_name="kilometre",
        display_name="km"
    )
    ```

    A unit that also references a physical dimension:

    ```
    Unit(
        odx_id=OdxLinkId("ID.kilometre", doc_frags),
        short_name="Kilometre",
        display_name="km",
        physical_dimension_ref=OdxLinkRef("ID.metre", doc_frags),
        factor_si_to_unit=1000,
        offset_si_to_unit=0
    )
    # where the physical_dimension_ref references, e.g.:
    PhysicalDimension(odx_id=OdxLinkId("ID.metre", doc_frags), short_name="metre", length_exp=1)
    ```
    """
    odx_id: OdxLinkId
    short_name: str
    display_name: str
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    factor_si_to_unit: Optional[float] = None
    offset_si_to_unit: Optional[float] = None
    physical_dimension_ref: Optional[OdxLinkRef] = None

    def __post_init__(self):
        self._physical_dimension = None

    @property
    def physical_dimension(self) -> PhysicalDimension:
        return self._physical_dimension

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        if self.physical_dimension_ref:
            self._physical_dimension = odxlinks.resolve(self.physical_dimension_ref)

            assert isinstance(self._physical_dimension, PhysicalDimension), (
                f"The physical_dimension_ref must be resolved to a PhysicalDimension."
                f" {self.physical_dimension_ref} referenced {self._physical_dimension}"
            )


@dataclass
class UnitGroup:
    """A group of units.

    There are two categories of groups: COUNTRY and EQUIV-UNITS.
    """
    short_name: str
    category: UnitGroupCategory
    unit_refs: List[OdxLinkRef] = field(default_factory=list)
    oid: Optional[str] = None
    long_name: Optional[str] = None
    description: Optional[str] = None

    def __post_init__(self):
        self._units = NamedItemList[Unit](short_name_as_id)

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        self._units = NamedItemList[Unit](
            short_name_as_id,
            [odxlinks.resolve(ref) for ref in self.unit_refs]
        )

    @property
    def units(self) -> NamedItemList[Unit]:
        return self._units


@dataclass
class UnitSpec:
    """
    A unit spec encapsulates three lists:

    * unit groups
    * units
    * physical_dimensions

    The following odx elements are not internalized: ADMIN-DATA, SDGS
    """
    # TODO (?): Why are there type errors...
    unit_groups: Union[NamedItemList[UnitGroup],
                       List[UnitGroup]] = field(default_factory=list)  # type: ignore
    units: Union[NamedItemList[Unit], List[Unit]] = field(
        default_factory=list)  # type: ignore
    physical_dimensions: Union[NamedItemList[PhysicalDimension],
                               List[PhysicalDimension]] = field(default_factory=list)  # type: ignore

    def __post_init__(self):
        self.unit_groups = NamedItemList(short_name_as_id, self.unit_groups)
        self.units = NamedItemList(short_name_as_id, self.units)
        self.physical_dimensions = NamedItemList(short_name_as_id, self.physical_dimensions)

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {}
        odxlinks.update({
            unit.odx_id: unit for unit in self.units
        })
        odxlinks.update({
            dim.odx_id: dim for dim in self.physical_dimensions
        })
        return odxlinks

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        for unit in self.units:
            unit._resolve_references(odxlinks)
        for group in self.unit_groups:
            group._resolve_references(odxlinks)


def read_unit_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    oid = et_element.get("OID")
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))
    display_name = et_element.findtext("DISPLAY-NAME")

    def read_optional_float(element, name):
        if element.findtext(name):
            return float(element.findtext(name))
        else:
            return None
    factor_si_to_unit = read_optional_float(et_element, "FACTOR-SI-TO-UNIT")
    offset_si_to_unit = read_optional_float(et_element, "OFFSET-SI-TO-UNIT")
    physical_dimension_ref = OdxLinkRef.from_et(et_element.find("PHYSICAL-DIMENSION-REF"), doc_frags)

    return Unit(
        odx_id=odx_id,
        short_name=short_name,
        display_name=display_name,
        oid=oid,
        long_name=long_name,
        description=description,
        factor_si_to_unit=factor_si_to_unit,
        offset_si_to_unit=offset_si_to_unit,
        physical_dimension_ref=physical_dimension_ref
    )


def read_physical_dimension_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    oid = et_element.get("OID")
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))

    def read_optional_int(element, name):
        if element.findtext(name):
            return int(element.findtext(name))
        else:
            return 0

    length_exp = read_optional_int(et_element, "LENGTH-EXP")
    mass_exp = read_optional_int(et_element, "MASS-EXP")
    time_exp = read_optional_int(et_element, "TIME-EXP")
    current_exp = read_optional_int(et_element, "CURRENT-EXP")
    temperature_exp = read_optional_int(et_element, "TEMPERATURE-EXP")
    molar_amount_exp = read_optional_int(et_element, "MOLAR-AMOUNT-EXP")
    luminous_intensity_exp = read_optional_int(et_element,
                                               "LUMINOUS-INTENSITY-EXP")

    return PhysicalDimension(
        odx_id=odx_id,
        short_name=short_name,
        oid=oid,
        long_name=long_name,
        description=description,
        length_exp=length_exp,
        mass_exp=mass_exp,
        time_exp=time_exp,
        current_exp=current_exp,
        temperature_exp=temperature_exp,
        molar_amount_exp=molar_amount_exp,
        luminous_intensity_exp=luminous_intensity_exp
    )


def read_unit_group_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    oid = et_element.get("OID")
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))
    category = et_element.findtext("CATEGORY")
    assert category in [
        "COUNTRY", "EQUIV-UNITS"], f'A UNIT-GROUP-CATEGORY must be "COUNTRY" or "EQUIV-UNITS". It was {category}.'
    unit_refs = []

    for el in et_element.iterfind("UNIT-REFS/UNIT-REF"):
        ref = OdxLinkRef.from_et(el, doc_frags)
        assert isinstance(ref, OdxLinkRef)
        unit_refs.append(ref)

    return UnitGroup(
        short_name=short_name,
        category=category,
        unit_refs=unit_refs,
        oid=oid,
        long_name=long_name,
        description=description
    )


def read_unit_spec_from_odx(et_element, doc_frags: List[OdxDocFragment]):

    unit_groups = [read_unit_group_from_odx(el, doc_frags)
                   for el in et_element.iterfind("UNIT-GROUPS/UNIT-GROUP")]
    units = [read_unit_from_odx(el, doc_frags)
             for el in et_element.iterfind("UNITS/UNIT")]
    physical_dimensions = [read_physical_dimension_from_odx(el, doc_frags)
                           for el in et_element.iterfind("PHYSICAL-DIMENSIONS/PHYSICAL-DIMENSION")]
    return UnitSpec(
        unit_groups=unit_groups,
        units=units,
        physical_dimensions=physical_dimensions
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from ..odxlink import OdxLinkDatabase
from .parameterbase import Parameter

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .diaglayer import DiagLayer

class TableKeyParameter(Parameter):
    def __init__(self,
                 short_name,
                 table_ref=None,
                 table_snref=None,
                 table_row_snref=None,
                 table_row_ref=None,
                 odx_id=None,
                 long_name=None,
                 byte_position=None,
                 bit_position=None,
                 semantic=None,
                 description=None):
        super().__init__(
            short_name=short_name,
            long_name=long_name,
            byte_position=byte_position,
            bit_position=bit_position,
            parameter_type="TABLE-KEY",
            semantic=semantic,
            description=description
        )
        self.table_ref = None
        self.table_snref = None
        self.table_row_ref = None
        self.table_row_snref = None
        if table_ref:
            self.table_ref = table_ref
            self.table_row_ref = table_row_ref
        elif table_snref:
            self.table_snref = table_snref
            self.table_row_snref = table_row_snref
        elif table_row_ref:
            self.table_row_ref = table_row_ref
        else:
            raise ValueError(
                "Either table_key_ref or table_key_snref must be defined.")
        self.odx_id = odx_id

    def is_required(self):
        raise NotImplementedError(
            "TableKeyParameter.is_required is not implemented yet.")

    def is_optional(self):
        raise NotImplementedError(
            "TableKeyParameter.is_optional is not implemented yet.")

    def get_coded_value(self):
        raise NotImplementedError(
            "Encoding a TableKeyParameter is not implemented yet.")

    def get_coded_value_as_bytes(self):
        raise NotImplementedError(
            "Encoding a TableKeyParameter is not implemented yet.")

    def decode_from_pdu(self, coded_message, default_byte_position=None):
        raise NotImplementedError(
            "Decoding a TableKeyParameter is not implemented yet.")

    def resolve_references(self,
                           parent_dl: "DiagLayer",
                           odxlinks: OdxLinkDatabase):
        self.table = None
        if self.table_snref:
            self.table = parent_dl.local_diag_data_dictionary_spec.tables[self.table_snref]
        elif self.table_ref:
            self.table = odxlinks.resolve(self.table_ref)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Dict, Literal, Optional, Any
from xml.etree import ElementTree

from .odxlink import OdxDocFragment

def read_description_from_odx(et_element: Optional[ElementTree.Element]) \
 -> Optional[str]:
    """Read a description tag.

    The description is located underneath the DESC tag of an an ODX
    element."""

    if et_element is None:
        return None

    if et_element.tag != "DESC":
        raise TypeError(f"Attempted to extract an ODX description from a "
                        f"'{et_element.tag}' XML node. (Must be a 'DESC' node!)")

    # Extract the contents of the tag as a XHTML string.
    raw_string = et_element.text or ''
    for e in et_element:
        raw_string += ElementTree.tostring(e, encoding='unicode')

    return raw_string.strip()


def short_name_as_id(obj: Any) -> str:
    """Retrieve an object's `short_name` attribute into a valid python identifier.

    Although short names are almost identical to python identifiers,
    their first character is allowed to be a number. This method
    prepends an underscore to such such shortnames.
    """

    sn = obj.short_name
    assert isinstance(sn, str)

    if sn[0].isdigit():
        return f"_{sn}"

    return sn





from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional
from xml.etree.ElementTree import Element

from .dataobjectproperty import DataObjectProperty, read_data_object_property_from_odx
from .nameditemlist import NamedItemList
from .odxlink import OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .units import UnitSpec, read_unit_spec_from_odx
from .admindata import AdminData, read_admin_data_from_odx
from .companydata import CompanyData, read_company_datas_from_odx
from .utils import read_description_from_odx, short_name_as_id


StandardizationLevel = Literal[
    "STANDARD",
    "OEM-SPECIFIC",
    "OPTIONAL",
    "OEM-OPTIONAL",
]

Usage = Literal[
    "ECU-SOFTWARE",
    "ECU-COMM",
    "APPLICATION",
    "TESTER",
]


def read_complex_value_from_odx(et_element):
    result = []
    for el in et_element.findall("*"):
        if el.tag == "SIMPLE-VALUE":
            result.append('' if el.text is None else el.text)
        else:
            result.append(read_complex_value_from_odx(el))
    return result


@dataclass
class BaseComparam:
    odx_id: OdxLinkId
    short_name: str
    long_name: Optional[str] = field(default=None, init=False)
    description: Optional[str] = field(default=None, init=False)
    physical_default_value: Any = field(default=None, init=False)
    param_class: str
    cptype: StandardizationLevel
    cpusage: Usage
    display_level: Optional[int] = field(default=None, init=False)

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        pass

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        return {self.odx_id: self}


@dataclass()
class ComplexComparam(BaseComparam):
    comparams: NamedItemList[BaseComparam]
    allow_multiple_values: Optional[bool] = None

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        for comparam in self.comparams:
            comparam._resolve_references(odxlinks)

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()
        for comparam in self.comparams:
            odxlinks.update(comparam._build_odxlinks())
        return odxlinks


@dataclass()
class Comparam(BaseComparam):
    dop_ref: OdxLinkRef
    _dop: Optional[DataObjectProperty] = field(default=None, init=False)

    @property
    def dop(self) -> DataObjectProperty:
        """The data object property describing this parameter."""
        assert self._dop is not None
        return self._dop

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        """Resolves the reference to the dop"""
        self._dop = odxlinks.resolve(self.dop_ref)


@dataclass()
class ComparamSubset:
    odx_id: Optional[OdxLinkId]
    short_name: str
    data_object_props: NamedItemList[DataObjectProperty]
    comparams: NamedItemList[BaseComparam]
    unit_spec: Optional[UnitSpec] = None
    long_name: Optional[str] = None
    description: Optional[str] = None
    admin_data: Optional[AdminData] = None
    company_datas: Optional[NamedItemList[CompanyData]] = None

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks: Dict[OdxLinkId, Any] = {}
        if self.odx_id is not None:
            odxlinks[self.odx_id] = self

        for dop in self.data_object_props:
            odxlinks[dop.odx_id] = dop

        for comparam in self.comparams:
            odxlinks.update(comparam._build_odxlinks())

        if self.unit_spec:
            odxlinks.update(self.unit_spec._build_odxlinks())

        return odxlinks

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        for dop in self.data_object_props:
            dop._resolve_references(odxlinks)

        for comparam in self.comparams:
            comparam._resolve_references(odxlinks)

        if self.unit_spec:
            self.unit_spec._resolve_references(odxlinks)


def read_comparam_from_odx(et_element, doc_frags: List[OdxDocFragment]) -> BaseComparam:
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    short_name = et_element.findtext("SHORT-NAME")
    param_class = et_element.attrib.get("PARAM-CLASS")
    cptype = et_element.attrib.get("CPTYPE")
    cpusage = et_element.attrib.get("CPUSAGE")

    comparam: BaseComparam
    if et_element.tag == "COMPARAM":
        dop_ref = OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), doc_frags)
        assert dop_ref is not None
        comparam = Comparam(
            odx_id=odx_id,
            short_name=short_name,
            param_class=param_class,
            cptype=cptype,
            cpusage=cpusage,
            dop_ref=dop_ref,
        )
        comparam.physical_default_value = et_element.findtext("PHYSICAL-DEFAULT-VALUE")
    elif et_element.tag == "COMPLEX-COMPARAM":
        comparams = [
            read_comparam_from_odx(el, doc_frags)
            for el in et_element if el.tag in ('COMPARAM', 'COMPLEX-COMPARAM')
        ]
        comparam = ComplexComparam(
            odx_id=odx_id,
            short_name=short_name,
            param_class=param_class,
            cptype=cptype,
            cpusage=cpusage,
            comparams=NamedItemList(short_name_as_id, comparams),
        )
        complex_values = et_element.iterfind("COMPLEX-PHYSICAL-DEFAULT-VALUE/COMPLEX-VALUES/COMPLEX-VALUE")
        comparam.physical_default_value = list(map(read_complex_value_from_odx, complex_values))
        tmp = et_element.get("ALLOW-MULTIPLE-VALUES")
        comparam.allow_multiple_values = (tmp == "true") if tmp is not None else None
    else:
        assert False, f"Failed to parse COMPARAM {short_name}"

    comparam.long_name = et_element.findtext("LONG-NAME")
    comparam.description = read_description_from_odx(et_element.find("DESC"))
    display_level = et_element.attrib.get("DISPLAY-LEVEL", None)
    if display_level is not None:
        comparam.display_level = int(display_level)

    return comparam


def read_comparam_subset_from_odx(et_element: Element) -> ComparamSubset:

    short_name = et_element.findtext("SHORT-NAME")
    assert short_name is not None

    doc_frags = [OdxDocFragment(short_name, str(et_element.tag))]
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))

    admin_data = \
        read_admin_data_from_odx(et_element.find("ADMIN-DATA"), doc_frags)
    company_datas = \
        read_company_datas_from_odx(et_element.find("COMPANY-DATAS"), doc_frags)

    data_object_props = [
        read_data_object_property_from_odx(el, doc_frags)
        for el in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
    ]
    comparams: List[BaseComparam] = []
    comparams += [
        read_comparam_from_odx(el, doc_frags)
        for el in et_element.iterfind("COMPARAMS/COMPARAM")
    ]
    comparams += [
        read_comparam_from_odx(el, doc_frags)
        for el in et_element.iterfind("COMPLEX-COMPARAMS/COMPLEX-COMPARAM")
    ]
    if et_element.find("UNIT-SPEC") is not None:
        unit_spec = read_unit_spec_from_odx(et_element.find("UNIT-SPEC"), doc_frags)
    else:
        unit_spec = None

    return ComparamSubset(
        odx_id=odx_id,
        short_name=short_name,
        long_name=long_name,
        description=description,
        admin_data=admin_data,
        company_datas=company_datas,
        data_object_props=NamedItemList(short_name_as_id, data_object_props),
        comparams=NamedItemList(short_name_as_id, comparams),
        unit_spec=unit_spec,
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from itertools import chain
from typing import Optional, Any, Dict, Iterable, List, Union
from copy import copy
import warnings
from xml.etree import ElementTree

from .utils import short_name_as_id
from .exceptions import DecodeError, OdxWarning
from .globals import logger, xsi
from .state import read_state_from_odx
from .state_transition import read_state_transition_from_odx

from .odxlink import OdxLinkRef, OdxLinkId, OdxLinkDatabase, OdxDocFragment
from .utils import read_description_from_odx
from .nameditemlist import NamedItemList
from .admindata import AdminData, read_admin_data_from_odx
from .companydata import CompanyData, read_company_datas_from_odx
from .communicationparameter import CommunicationParameterRef, read_communication_param_ref_from_odx
from .diagdatadictionaryspec import DiagDataDictionarySpec, read_diag_data_dictionary_spec_from_odx
from .dataobjectproperty import DopBase
from .functionalclass import read_functional_class_from_odx
from .audience import read_additional_audience_from_odx
from .message import Message
from .service import DiagService, read_diag_service_from_odx
from .singleecujob import SingleEcuJob, read_single_ecu_job_from_odx
from .structures import Request, Response, read_structure_from_odx

# Defines priority of overriding objects
PRIORITY_OF_DIAG_LAYER_TYPE = {
    "PROTOCOL": 1,
    "FUNCTIONAL-GROUP": 2,
    "BASE-VARIANT": 3,
    "ECU-VARIANT": 4,
    # Inherited services from ECU Shared Data always override inherited services from other diag layers
    "ECU-SHARED-DATA": 5
}


class DiagLayer:

    class ParentRef:
        def __init__(self,
                     parent: Union[OdxLinkRef, "DiagLayer"],
                     ref_type: str,
                     not_inherited_diag_comms=[],
                     not_inherited_dops=[]):
            """
            Parameters
            ----------
            parent: OdxLinkRef | DiagLayer
                A reference to the or the parent DiagLayer
            ref_type: str
            not_inherited_diag_comms: List[str]
                short names of not inherited diag comms
            not_inherited_dops: List[str]
                short names of not inherited DOPs
            """
            if ref_type not in ["PROTOCOL-REF", "BASE-VARIANT-REF",
                                "ECU-SHARED-DATA-REF", "FUNCTIONAL-GROUP-REF"]:
                warnings.warn(f'Unknown parent ref type {ref_type}', OdxWarning)
            if isinstance(parent, OdxLinkRef):
                self.parent_ref = parent
                self.parent_diag_layer = None
            else:
                assert isinstance(parent, DiagLayer)

                self.parent_ref = OdxLinkRef.from_id(parent.odx_id)
                self.parent_diag_layer = parent
            self.not_inherited_diag_comms = not_inherited_diag_comms
            self.not_inherited_dops = not_inherited_dops
            self.ref_type = ref_type

        def _resolve_references(self, odxlinks: OdxLinkDatabase):
            self.parent_diag_layer = odxlinks.resolve(self.parent_ref)

        def get_inheritance_priority(self):
            return PRIORITY_OF_DIAG_LAYER_TYPE[self.parent_diag_layer.variant_type]

        def get_inherited_services_by_name(self):
            services = {service.short_name: service for service in self.parent_diag_layer._services
                        if service.short_name not in self.not_inherited_diag_comms}
            return services

        def get_inherited_data_object_properties_by_name(self):
            dops = {dop.short_name: dop for dop in self.parent_diag_layer._data_object_properties
                    if dop.short_name not in self.not_inherited_dops}
            return dops

        def get_inherited_communication_parameters_by_name(self):
            return {cp.id_ref.ref_id: cp for cp in self.parent_diag_layer._communication_parameters}

    def __init__(self,
                 variant_type,
                 odx_id,
                 short_name,
                 long_name=None,
                 description=None,
                 requests: List[Request] = [],
                 positive_responses: List[Response] = [],
                 negative_responses: List[Response] = [],
                 services: List[DiagService] = [],
                 single_ecu_jobs: List[SingleEcuJob] = [],
                 diag_comm_refs: List[OdxLinkRef] = [],
                 parent_refs: List[ParentRef] = [],
                 diag_data_dictionary_spec: Optional[DiagDataDictionarySpec] = None,
                 communication_parameters:
                 Iterable[CommunicationParameterRef] = [],
                 odxlinks=None,
                 additional_audiences=[],
                 functional_classes=[],
                 states=[],
                 state_transitions=[],
                 import_refs=[],
                 ):
        logger.info(f"Initializing variant type {variant_type}")
        self.variant_type = variant_type

        self.odx_id = odx_id
        self.short_name = short_name
        self.long_name = long_name
        self.description = description

        # Requests and Responses
        self.requests = requests
        self.positive_responses = NamedItemList[Response](short_name_as_id,
                                                          positive_responses)
        self.negative_responses = NamedItemList[Response](short_name_as_id,
                                                          negative_responses)

        # ParentRefs
        self.parent_refs = parent_refs

        # DiagServices (note that they do not include inherited services!)
        self._local_services = NamedItemList[DiagService](short_name_as_id,
                                                          services)
        self._local_single_ecu_jobs = NamedItemList[SingleEcuJob](short_name_as_id,
                                                                  single_ecu_jobs)
        self._diag_comm_refs = diag_comm_refs

        # DOP-BASEs
        self.local_diag_data_dictionary_spec = diag_data_dictionary_spec

        # Communication parameters, e.g. CAN-IDs
        self._local_communication_parameters = communication_parameters

        self.additional_audiences = additional_audiences
        self.functional_classes = functional_classes
        self.states = states
        self.state_transitions = state_transitions

        # Properties that include inherited objects
        self._services: NamedItemList[Union[DiagService, SingleEcuJob]]\
            = NamedItemList(short_name_as_id, [])
        self._communication_parameters: NamedItemList[CommunicationParameterRef]\
            = NamedItemList(short_name_as_id, [])
        self._data_object_properties: NamedItemList[DopBase]\
            = NamedItemList(short_name_as_id, [])

        self.import_refs = import_refs

        if odxlinks is not None:
            self.finalize_init(odxlinks)

    @property
    def services(self) -> NamedItemList[Union[DiagService, SingleEcuJob]]:
        """All services that this diagnostic layer offers including inherited services."""
        return self._services

    @property
    def data_object_properties(self) -> NamedItemList[DopBase]:
        """All data object properties including inherited ones.
        This attribute corresponds to all specializations of DOP-BASE
        defined in the DIAG-DATA-DICTIONARY-SPEC of this diag layer as well as
        in the DIAG-DATA-DICTIONARY-SPEC of any parent.
        """
        return self._data_object_properties

    @property
    def communication_parameters(self) -> NamedItemList[CommunicationParameterRef]:
        """All communication parameters including inherited ones."""
        return self._communication_parameters

    def finalize_init(self, odxlinks: Optional[OdxLinkDatabase] = None):
        """Resolves all references.

        This method should be called whenever the diag layer (or a referenced object) was changed.
        Particularly, this method assumes that all inherited diag layer are correctly initialized,
        i.e., have resolved their references.
        """

        if odxlinks is None:
            odxlinks = OdxLinkDatabase()

        odxlinks.update(self._build_odxlinks())
        self._resolve_references(odxlinks)

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        """Construct a mapping from IDs to all objects that are contained in this diagnostic layer."""
        logger.info(f"Adding {self.odx_id} to odxlinks.")

        odxlinks = {}

        for obj in chain(self._local_services,
                         self._local_single_ecu_jobs,
                         self.requests,
                         self.positive_responses,
                         self.negative_responses,
                         self.additional_audiences,
                         self.functional_classes,
                         self.states,
                         self.state_transitions):
            odxlinks[obj.odx_id] = obj

        if self.local_diag_data_dictionary_spec:
            odxlinks.update(
                self.local_diag_data_dictionary_spec._build_odxlinks()
            )

        odxlinks[self.odx_id] = self
        return odxlinks

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        """Recursively resolve all references."""
        # Resolve inheritance
        for pr in self.parent_refs:
            pr._resolve_references(odxlinks)

        services = sorted(self._compute_available_services_by_name(odxlinks).values(),
                          key=short_name_as_id)
        self._services = NamedItemList[Union[DiagService, SingleEcuJob]](
            short_name_as_id,
            services)

        dops = sorted(self._compute_available_data_object_properties_by_name().values(),
                      key=short_name_as_id)
        self._data_object_properties = NamedItemList[DopBase](
            short_name_as_id,
            dops)
        for comparam in self._local_communication_parameters:
            comparam._resolve_references(odxlinks)

        self._communication_parameters = NamedItemList[CommunicationParameterRef](
            short_name_as_id,
             list(self._compute_available_commmunication_parameters_by_name().values())
        )

        # Resolve all other references
        for struct in chain(self.requests,
                            self.positive_responses,
                            self.negative_responses):
            struct._resolve_references(self, odxlinks)

        local_diag_comms: Iterable[Union[DiagService, SingleEcuJob]] \
            = (*self._local_services, *self._local_single_ecu_jobs)
        for service in local_diag_comms:
            service._resolve_references(odxlinks)

        if self.local_diag_data_dictionary_spec:
            self.local_diag_data_dictionary_spec._resolve_references(self,
                                                                     odxlinks)

    def __local_services_by_name(self, odxlinks: OdxLinkDatabase) -> Dict[str, Union[DiagService, SingleEcuJob]]:
        services_by_name: Dict[str, Union[DiagService, SingleEcuJob]] = {}

        for ref in self._diag_comm_refs:
            if (obj := odxlinks.resolve_lenient(ref)) is not None:
                services_by_name[obj.short_name] = obj
            else:
                logger.warning(f"Diag comm ref {ref!r} could not be resolved.")

        services_by_name.update({
            service.short_name: service for service in self._local_services
        })
        services_by_name.update({
            service.short_name: service for service in self._local_single_ecu_jobs
        })
        return services_by_name

    def _compute_available_services_by_name(self, odxlinks: OdxLinkDatabase) -> Dict[str, DiagService]:
        """Helper method for initializing the available services.
        This computes the services that are inherited from other diagnostic layers."""
        services_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            services_by_name.update(
                parent_ref.get_inherited_services_by_name())

        services_by_name.update(self.__local_services_by_name(odxlinks))
        return services_by_name

    def _compute_available_data_object_properties_by_name(self) -> Dict[str, DopBase]:
        """Returns the locally defined and inherited DOPs."""
        data_object_properties_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            data_object_properties_by_name.update(
                parent_ref.get_inherited_data_object_properties_by_name())

        if self.local_diag_data_dictionary_spec:
            data_object_properties_by_name.update({
                d.short_name: d for d in self.local_diag_data_dictionary_spec.all_data_object_properties
            })
        return data_object_properties_by_name

    def _compute_available_commmunication_parameters_by_name(self) -> dict:
        com_params_by_name = {}

        # Look in parent refs for inherited services
        # Fetch services from low priority parents first, then update with increasing priority
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            com_params_by_name.update(
                parent_ref.get_inherited_communication_parameters_by_name())

        com_params_by_name.update(
            {cp.id_ref.ref_id: cp for cp in self._local_communication_parameters})
        return com_params_by_name

    def _get_parent_refs_sorted_by_priority(self, reverse=False):
        return sorted(self.parent_refs, key=lambda pr: pr.get_inheritance_priority(), reverse=reverse)

    def _build_coded_prefix_tree(self):
        """Constructs the coded prefix tree of the services.
        Each leaf node is a list of `DiagService`s.
        (This is because navigating from a service to the request/ responses is easier than finding the service for a given request/response object.)

        Example:
        Let there be four services with corresponding requests:
        * Request 1 has the coded constant prefix `12 34`.
        * Request 2 has the coded constant prefix `12 34`.
        * Request 3 has the coded constant prefix `12 56`.
        * Request 4 has the coded constant prefix `12 56 00`.

        Then, the constructed prefix tree is the dict
        ```
        {0x12: {0x34: {-1: [<Service 1>, <Service 2>]},
                0x56: {-1: [<Service 3>],
                       0x0: {-1: [<Service 4>]}
                       }}}
        ```
        Note, that the inner `-1` are constant to distinguish them from possible service IDs.

        Also note, that it is actually allowed that
        (a) SIDs for different services are the same like for service 1 and 2 (thus each leaf node is a list) and
        (b) one SID is the prefix of another SID like for service 3 and 4 (thus the constant `-1` key).
        """
        services = [s for s in self._services if isinstance(s, DiagService)]
        prefix_tree = {}
        for s in services:
            # Compute prefixes for the request and all responses
            request_prefix = s.request.coded_const_prefix()
            prefixes = [request_prefix] + [
                message.coded_const_prefix(
                    request_prefix=request_prefix
                ) for message in chain(s.positive_responses, s.negative_responses)
            ]
            for coded_prefix in prefixes:
                # Traverse prefix tree
                sub_tree = prefix_tree
                for b in coded_prefix:
                    if sub_tree.get(b) is None:
                        sub_tree[b] = {}
                    sub_tree = sub_tree.get(b)

                    assert isinstance(
                        sub_tree, dict), f"{sub_tree} has type {type(sub_tree)}. How did this happen?"
                # Add service as leaf to prefix tree
                if sub_tree.get(-1) is None:
                    sub_tree[-1] = [s]
                else:
                    sub_tree[-1].append(s)
        return prefix_tree

    def _find_services_for_uds(self, message: Union[bytes, bytearray]):
        if not hasattr(self, "_prefix_tree"):
            # Compute the prefix tree the first time this decode function is called.
            self._prefix_tree = self._build_coded_prefix_tree()
        prefix_tree = self._prefix_tree

        # Find matching service(s) in prefix tree
        possible_services = []
        for b in message:
            if prefix_tree.get(b) is not None:
                assert isinstance(prefix_tree.get(b), dict)
                prefix_tree = prefix_tree.get(b)
            else:
                break
            if -1 in prefix_tree:
                possible_services += prefix_tree[-1]
        return possible_services

    def decode(self, message: Union[bytes, bytearray]) -> Iterable[Message]:
        possible_services = self._find_services_for_uds(message)

        if possible_services is None:
            raise DecodeError(
                f"Couldn't find corresponding service for message {message.hex()}."
            )

        decoded_messages = []

        for service in possible_services:
            try:
                decoded_messages.append(service.decode_message(message))
            except DecodeError as e:
                pass
        if len(decoded_messages) == 0:
            raise DecodeError(
                f"None of the services {possible_services} could parse {message.hex()}.")
        return decoded_messages

    def decode_response(
            self, response: Union[bytes, bytearray],
            request: Union[bytes, bytearray, Message]) -> Iterable[Message]:
        if isinstance(request, Message):
            possible_services = [request.service]
        else:
            if not isinstance(request, (bytes, bytearray)):
                raise TypeError(f"Request parameter must have type "
                                f"Message, bytes or bytearray but was {type(request)}")
            possible_services = self._find_services_for_uds(request)
        if possible_services is None:
            raise DecodeError(
                f"Couldn't find corresponding service for request {request.hex()}."
            )

        decoded_messages = []

        for service in possible_services:
            try:
                decoded_messages.append(service.decode_message(response))
            except DecodeError as e:
                pass
        if len(decoded_messages) == 0:
            raise DecodeError(
                f"None of the services {possible_services} could parse {response.hex()}.")
        return decoded_messages

    def get_communication_parameter(self, name: str) \
            -> Optional[CommunicationParameterRef]:

        cps = [cp for cp in self.communication_parameters if cp.short_name == name]
        if len(cps) > 1:
            warnings.warn(f"Communication parameter `{name}` specified more "
                          f"than once. Using first occurence.", OdxWarning)
        elif len(cps) == 0:
            return None

        return cps[0]

    def get_receive_id(self) -> Optional[int]:
        """CAN ID to which the ECU listens for diagnostic messages"""
        com_param = self.get_communication_parameter("CP_UniqueRespIdTable")
        if com_param is None:
            return None
        value = com_param.value
        if not isinstance(value, dict):
            return None

        return value.get('CP_CanPhysReqId', None)

    def get_send_id(self) -> Optional[int]:
        """CAN ID to which the ECU sends replies to diagnostic messages"""
        com_param = self.get_communication_parameter("CP_UniqueRespIdTable")
        if com_param is None:
            return None
        value = com_param.value
        if not isinstance(value, dict):
            return None

        return value.get('CP_CanRespUSDTId', None)

    def get_can_func_req_id(self) -> Optional[int]:
        """CAN Functional Request Id."""
        com_param = self.get_communication_parameter("CP_CanFuncReqId")

        if com_param is None:
            return None

        return int(com_param.value)

    def get_logical_doip_address(self) -> Optional[int]:
        """The logical DoIP address of the ECU."""
        com_param = self.get_communication_parameter("CP_UniqueRespIdTable")

        if com_param is None:
            return None

        value = com_param.value
        if not isinstance(value, dict):
            return None

        return value.get('CP_CanPhysReqExtAddr', None)

    def get_tester_present_time(self) -> Optional[float]:
        """Timeout on inactivity in seconds.

        This is defined by the communication parameter "CP_TesterPresentTime".
        If the variant does not define this parameter, the default value 3.0 is returned.

        Description of the comparam: "Time between a response and the next subsequent tester present message
        (if no other request is sent to this ECU) in case of physically addressed requests."
        """
        com_param = self.get_communication_parameter("CP_TesterPresentTime")

        if com_param is None:
            return 3.0  # default specified by the standard

        return int(com_param.value) / 1e6

    def __repr__(self) -> str:
        return f"""DiagLayer(variant_type={self.variant_type},
          odx_id={repr(self.odx_id)},
          short_name={repr(self.short_name)},
          long_name={repr(self.long_name)},
          description={repr(self.description)},
          requests={self.requests},
          positive_responses={self.positive_responses},
          negative_responses={self.negative_responses},
          services={self._local_services},
          diag_comm_refs={self._diag_comm_refs},
          parent_refs={self.parent_refs},
          diag_data_dictionary_spec={self.local_diag_data_dictionary_spec},
          communication_parameters={self._local_communication_parameters})"""

    def __str__(self) -> str:
        return f"DiagLayer('{self.short_name}', type='{self.variant_type}')"


def read_parent_ref_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
        -> DiagLayer.ParentRef:
    parent_ref = OdxLinkRef.from_et(et_element, doc_frags)
    assert parent_ref is not None

    not_inherited_diag_comms = [el.get("SHORT-NAME")
                                for el in et_element.iterfind(
                                    "NOT-INHERITED-DIAG-COMMS/NOT-INHERITED-DIAG-COMM/DIAG-COMM-SNREF")]
    not_inherited_dops = [el.get("SHORT-NAME")
                          for el in et_element.iterfind("NOT-INHERITED-DOPS/NOT-INHERITED-DOP/DOP-BASE-SNREF")]
    ref_type = et_element.get(f"{xsi}type")

    return DiagLayer.ParentRef(
        parent_ref,
        ref_type=ref_type,
        not_inherited_diag_comms=not_inherited_diag_comms,
        not_inherited_dops=not_inherited_dops
    )


def read_diag_layer_from_odx(et_element: ElementTree.Element,
                             doc_frags: List[OdxDocFragment]) \
        -> DiagLayer:

    variant_type = et_element.tag

    short_name = et_element.findtext("SHORT-NAME")
    assert short_name is not None
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))

    logger.info(f"Parsing diagnostic layer '{short_name}' "
                f"of type {variant_type} ...")

    # extend the applicable ODX "document fragments" for the diag layer objects
    doc_frags = copy(doc_frags)
    doc_frags.append(OdxDocFragment(short_name, "LAYER"))

    odx_id = OdxLinkId.from_et(et_element, doc_frags)

    # Parse DiagServices
    services = [read_diag_service_from_odx(service, doc_frags)
                for service in et_element.iterfind("DIAG-COMMS/DIAG-SERVICE")]
    diag_comm_refs = []
    for service in et_element.iterfind("DIAG-COMMS/DIAG-COMM-REF"):
        ref = OdxLinkRef.from_et(service, doc_frags)
        assert ref is not None
        diag_comm_refs.append(ref)

    single_ecu_jobs = [read_single_ecu_job_from_odx(sej, doc_frags)
                       for sej in et_element.iterfind("DIAG-COMMS/SINGLE-ECU-JOB")]

    # Parse ParentRefs
    parent_refs = [read_parent_ref_from_odx(pr_el, doc_frags)
                   for pr_el in et_element.iterfind("PARENT-REFS/PARENT-REF")]

    # Parse communication parameter refs
    com_params = [read_communication_param_ref_from_odx(el, doc_frags)
                  for el in et_element.iterfind("COMPARAM-REFS/COMPARAM-REF")]

    # Parse Requests and Responses
    requests = []
    for rq_elem in et_element.iterfind("REQUESTS/REQUEST"):
        rq = read_structure_from_odx(rq_elem, doc_frags)
        assert isinstance(rq, Request)
        requests.append(rq)

    positive_responses = []
    for pr_elem in et_element.iterfind("POS-RESPONSES/POS-RESPONSE"):
        pr = read_structure_from_odx(pr_elem, doc_frags)
        assert isinstance(pr, Response)
        positive_responses.append(pr)

    negative_responses = []
    for nr_elem in et_element.iterfind("NEG-RESPONSES/NEG-RESPONSE"):
        nr = read_structure_from_odx(nr_elem, doc_frags)
        assert isinstance(nr, Response)
        negative_responses.append(nr)

    additional_audiences = [read_additional_audience_from_odx(el, doc_frags)
                            for el in et_element.iterfind("ADDITIONAL-AUDIENCES/ADDITIONAL-AUDIENCE")]

    functional_classes = [
        read_functional_class_from_odx(el, doc_frags)
        for el in et_element.iterfind("FUNCT-CLASSS/FUNCT-CLASS")]

    states = [
        read_state_from_odx(el, doc_frags)
        for el in et_element.iterfind("STATE-CHARTS/STATE-CHART/STATES/STATE")]

    state_transitions = [
        read_state_transition_from_odx(el, doc_frags)
        for el in et_element.iterfind("STATE-CHARTS/"
                                      "STATE-CHART/"
                                      "STATE-TRANSITIONS/"
                                      "STATE-TRANSITION")]

    if et_element.find("DIAG-DATA-DICTIONARY-SPEC"):
        diag_data_dictionary_spec = read_diag_data_dictionary_spec_from_odx(
            et_element.find("DIAG-DATA-DICTIONARY-SPEC"), doc_frags)
    else:
        diag_data_dictionary_spec = None

    import_refs = [OdxLinkRef.from_et(ref, doc_frags)
                   for ref in et_element.iterfind("IMPORT-REFS/IMPORT-REF")]

    # TODO: Are UNIT-SPEC and SDGS needed?

    # Create DiagLayer
    dl = DiagLayer(variant_type,
                   odx_id,
                   short_name,
                   long_name=long_name,
                   description=description,
                   requests=requests,
                   positive_responses=positive_responses,
                   negative_responses=negative_responses,
                   services=services,
                   diag_comm_refs=diag_comm_refs,
                   single_ecu_jobs=single_ecu_jobs,
                   parent_refs=parent_refs,
                   diag_data_dictionary_spec=diag_data_dictionary_spec,
                   communication_parameters=com_params,
                   additional_audiences=additional_audiences,
                   functional_classes=functional_classes,
                   states=states,
                   state_transitions=state_transitions,
                   import_refs=import_refs,
                   )

    return dl

class DiagLayerContainer:
    def __init__(self,
                 odx_id: OdxLinkId,
                 short_name: str,
                 long_name: Optional[str] = None,
                 description: Optional[str] = None,
                 admin_data: Optional[AdminData] = None,
                 company_datas: Optional[NamedItemList[CompanyData]] = None,
                 ecu_shared_datas: List[DiagLayer] = [],
                 protocols: List[DiagLayer] = [],
                 functional_groups: List[DiagLayer] = [],
                 base_variants: List[DiagLayer] = [],
                 ecu_variants: List[DiagLayer] = []
                 ) -> None:
        self.odx_id = odx_id
        self.short_name = short_name
        self.long_name = long_name
        self.description = description
        self.admin_data = admin_data
        self.company_datas = company_datas

        self.ecu_shared_datas = ecu_shared_datas
        self.protocols = protocols
        self.functional_groups = functional_groups
        self.base_variants = base_variants
        self.ecu_variants = ecu_variants

        self._diag_layers = NamedItemList[DiagLayer](short_name_as_id, list(
            chain(self.ecu_shared_datas, self.protocols, self.functional_groups, self.base_variants, self.ecu_variants)))

    def _build_odxlinks(self):
        result = {}
        result[self.odx_id] = self

        if self.admin_data is not None:
            result.update(self.admin_data._build_odxlinks())

        if self.company_datas is not None:
            for cd in self.company_datas:
                result.update(cd._build_odxlinks())

        return result

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_references(odxlinks)

        if self.company_datas is not None:
            for cd in self.company_datas:
                cd._resolve_references(odxlinks)

    @property
    def diag_layers(self):
        return self._diag_layers

    def __getitem__(self, key: Union[int, str]) -> DiagLayer:
        return self.diag_layers[key]

    def __repr__(self) -> str:
        return f"DiagLayerContainer('{self.short_name}')"

    def __str__(self) -> str:
        return f"DiagLayerContainer('{self.short_name}')"


def read_diag_layer_container_from_odx(et_element):
    short_name = et_element.findtext("SHORT-NAME")
    assert short_name is not None
    long_name = et_element.findtext("LONG-NAME")

    # create the current ODX "document fragment" (description of the
    # current document for references and IDs)
    doc_frags = [OdxDocFragment(short_name, "CONTAINER")]

    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    description = read_description_from_odx(et_element.find("DESC"))
    admin_data = read_admin_data_from_odx(et_element.find("ADMIN-DATA"), doc_frags)
    company_datas = read_company_datas_from_odx(et_element.find("COMPANY-DATAS"), doc_frags)
    ecu_shared_datas = [read_diag_layer_from_odx(dl_element,
                                                 doc_frags)
                        for dl_element in et_element.iterfind("ECU-SHARED-DATAS/ECU-SHARED-DATA")]
    protocols = [read_diag_layer_from_odx(dl_element,
                                          doc_frags)
                 for dl_element in et_element.iterfind("PROTOCOLS/PROTOCOL")]
    functional_groups = [read_diag_layer_from_odx(dl_element,
                                                  doc_frags)
                         for dl_element in et_element.iterfind("FUNCTIONAL-GROUPS/FUNCTIONAL-GROUP")]
    base_variants = [read_diag_layer_from_odx(dl_element,
                                              doc_frags)
                     for dl_element in et_element.iterfind("BASE-VARIANTS/BASE-VARIANT")]
    ecu_variants = [read_diag_layer_from_odx(dl_element,
                                             doc_frags)
                    for dl_element in et_element.iterfind("ECU-VARIANTS/ECU-VARIANT")]

    return DiagLayerContainer(odx_id,
                              short_name,
                              long_name=long_name,
                              description=description,
                              admin_data=admin_data,
                              company_datas=company_datas,
                              ecu_shared_datas=ecu_shared_datas,
                              protocols=protocols,
                              functional_groups=functional_groups,
                              base_variants=base_variants,
                              ecu_variants=ecu_variants
                              )



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from ..load_file import load_file as _load_file


def add_pdx_argument(parser):
    parser.add_argument(
        "pdx_file", metavar="PDX_FILE", help="path to the .pdx file")
    #parser.add_argument("pdx_files", metavar="PDX_FILES", nargs="+", help="PDX descriptions of all ECUs which shall be analyzed")


def load_file(args):
    db_file_name = args.pdx_file
    odxdb = None
    if db_file_name is not None:
        odxdb = _load_file(db_file_name)
    return odxdb



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from .nameditemlist import NamedItemList
from .utils import read_description_from_odx
from .odxlink import OdxLinkId, OdxLinkDatabase, OdxDocFragment
from .utils import short_name_as_id

from dataclasses import dataclass, field
from typing import Optional, Any, Dict, List

@dataclass()
class XDoc:
    short_name : str
    long_name : Optional[str] = None
    description : Optional[str] = None
    number: Optional[str] = None
    state: Optional[str] = None
    date: Optional[str] = None
    publisher: Optional[str] = None
    url: Optional[str] = None
    position: Optional[str] = None

@dataclass()
class RelatedDoc:
    description: Optional[str] = None
    xdoc: Optional[XDoc] = None

@dataclass()
class CompanySpecificInfo:
    related_docs: Optional[List[RelatedDoc]]

@dataclass()
class TeamMember:
    odx_id: OdxLinkId
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    department: Optional[str] = None
    address: Optional[str] = None
    zip: Optional[str] = None
    city: Optional[str] = None
    phone: Optional[str] = None
    fax: Optional[str] = None
    email: Optional[str] = None

@dataclass()
class CompanyData:
    odx_id: OdxLinkId
    short_name: str
    long_name: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    team_members: Optional[NamedItemList[TeamMember]] = None
    company_specific_info: Optional[CompanySpecificInfo] = None

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        result = { self.odx_id: self }

        # team members
        if self.team_members is not None:
            for tm in self.team_members:
                result[tm.odx_id] = tm

        return result

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        pass

def read_xdoc_from_odx(xdoc):
    short_name = xdoc.findtext("SHORT-NAME")
    long_name = xdoc.findtext("LONG-NAME")
    description = read_description_from_odx(xdoc.find("DESC"))
    number = xdoc.findtext("NUMBER")
    state = xdoc.findtext("STATE")
    date = xdoc.findtext("DATE")
    publisher = xdoc.findtext("PUBLISHER")
    url = xdoc.findtext("URL")
    position = xdoc.findtext("POSITION")

    return XDoc(short_name=short_name,
                long_name=long_name,
                description=description,
                number=number,
                state=state,
                date=date,
                publisher=publisher,
                url=url,
                position=position,
                )

def read_company_datas_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
        -> NamedItemList[CompanyData]:
    if et_element is None:
        return NamedItemList(short_name_as_id)

    cdl = NamedItemList(short_name_as_id) # type: ignore

    for cd in et_element.iterfind("COMPANY-DATA"):
        odx_id = OdxLinkId.from_et(cd, doc_frags)
        assert odx_id is not None
        short_name = cd.findtext("SHORT-NAME")
        long_name = cd.findtext("LONG-NAME")
        description = read_description_from_odx(cd.find("DESC"))
        roles = cd.find("ROLES")
        if roles is not None:
            rlist = list()

            for role in roles.iterfind("ROLE"):
                rlist.append(role.text)

            roles = rlist

        team_members = cd.find("TEAM-MEMBERS")
        if team_members is not None:
            tml = NamedItemList(short_name_as_id) # type: ignore

            for tm in team_members.iterfind("TEAM-MEMBER"):
                tm_id = OdxLinkId.from_et(tm, doc_frags)
                assert tm_id is not None
                tm_short_name = tm.findtext("SHORT-NAME")
                tm_long_name = tm.findtext("LONG-NAME")
                tm_description = read_description_from_odx(tm.find("DESC"))

                tm_roles = tm.find("ROLES")
                if tm_roles is not None:
                    rlist = list()
                    for role in tm_roles.iterfind("ROLE"):
                        rlist.append(role.text)
                    tm_roles = rlist

                tm_department = tm.findtext("DEPARTMENT")
                tm_address = tm.findtext("ADDRESS")
                tm_zip = tm.findtext("ZIP")
                tm_city = tm.findtext("CITY")
                tm_phone = tm.findtext("PHONE")
                tm_fax = tm.findtext("FAX")
                tm_email = tm.findtext("EMAIL")

                tml.append(TeamMember(odx_id=tm_id,
                                      short_name=tm_short_name,
                                      long_name=tm_long_name,
                                      description=tm_description,
                                      roles=tm_roles,
                                      department=tm_department,
                                      address=tm_address,
                                      zip=tm_zip,
                                      city=tm_city,
                                      phone=tm_phone,
                                      fax=tm_fax,
                                      email=tm_email))

            team_members = tml

        company_specific_info = cd.find("COMPANY-SPECIFIC-INFO")
        if company_specific_info is not None:
            related_docs = company_specific_info.find("RELATED-DOCS")
            if related_docs is not None:
                rdlist = list()
                for rd in related_docs.iterfind("RELATED-DOC"):
                    rd_description = read_description_from_odx(rd.find("DESC"))
                    xdoc = rd.find("XDOC")
                    if xdoc is not None:
                        xdoc = read_xdoc_from_odx(xdoc)

                    rdlist.append(RelatedDoc(description=rd_description,
                                             xdoc=xdoc,
                                             ))

                related_docs = rdlist

            company_specific_info = CompanySpecificInfo(related_docs=related_docs)

        cdl.append(CompanyData(odx_id=odx_id,
                               short_name=short_name,
                               long_name=long_name,
                               description=description,
                               roles=roles,
                               team_members=team_members,
                               company_specific_info=company_specific_info))

    return cdl



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from typing import Optional, List, TYPE_CHECKING
import warnings

from odxtools.exceptions import OdxWarning

from .odxlink import OdxLinkDatabase, OdxLinkRef, OdxDocFragment
from .utils import read_description_from_odx
from .comparam_subset import BaseComparam, Comparam, ComplexComparam, read_complex_value_from_odx


def _get_comparam_value(comparam: BaseComparam, value):
    if not comparam:
        return value
    if value is None:
        value = comparam.physical_default_value

    if isinstance(comparam, Comparam):
        return comparam.dop.physical_type.base_data_type.from_string(value)

    if isinstance(comparam, ComplexComparam) and isinstance(value, list):
        value = value + [None] * (len(comparam.comparams) - len(value))
        result = dict()
        for i, cp in enumerate(comparam.comparams):
            result[cp.short_name] = _get_comparam_value(cp, value[i])
        return result

    return value


class CommunicationParameterRef:

    def __init__(self,
                 value,
                 id_ref: OdxLinkRef,
                 description: Optional[str] = None,
                 protocol_sn_ref: Optional[str] = None,
                 prot_stack_sn_ref: Optional[str] = None):
        self._value = value
        self.id_ref = id_ref
        self.description = description
        self.protocol_sn_ref = protocol_sn_ref
        self.prot_stack_sn_ref = prot_stack_sn_ref
        self.comparam: Optional[BaseComparam] = None

    def __repr__(self) -> str:
        val = self.value
        if isinstance(val, str):
            val = f"'{val}'"
        return f"CommunicationParameter('{self.short_name}', value={val})"

    def __str__(self) -> str:
        return self.__repr__()

    def _resolve_references(self, odxlinks: OdxLinkDatabase):
        # Temporary lenient until tests are updated
        self.comparam = odxlinks.resolve_lenient(self.id_ref)
        if not self.comparam:
            warnings.warn(f"Could not resolve COMPARAM '{self.id_ref}'", OdxWarning)

    @property
    def short_name(self):
        if self.comparam:
            return self.comparam.short_name
        # ODXLINK IDs allow dots and hyphens, but python identifiers do not.
        # This should not happen anyway in a correct PDX
        return self.id_ref.ref_id.replace(".", "__").replace("-", "_")

    @property
    def value(self):
        return _get_comparam_value(self.comparam, self._value)


def read_communication_param_ref_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    id_ref = OdxLinkRef.from_et(et_element, doc_frags)
    assert id_ref is not None

    # ODX standard v2.0.0 defined only VALUE
    # ODX standard v2.0.1 decided to break things and change it to choice between SIMPLE-VALUE and COMPLEX-VALUE
    if et_element.find("VALUE") is not None:
        value = et_element.findtext("VALUE")
    elif et_element.find("SIMPLE-VALUE") is not None:
        value = et_element.findtext("SIMPLE-VALUE")
    else:
        value = read_complex_value_from_odx(et_element.find("COMPLEX-VALUE"))

    description = read_description_from_odx(et_element.find("DESC"))
    protocol_sn_ref = et_element.findtext("PROTOCOL-SNREF")
    prot_stack_sn_ref = et_element.findtext("PROT-STACK-SNREF")
    return CommunicationParameterRef(value,
                                     id_ref,
                                     description=description,
                                     protocol_sn_ref=protocol_sn_ref,
                                     prot_stack_sn_ref=prot_stack_sn_ref
                                     )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

import typing
import warnings

from dataclasses import dataclass, field
from xml.etree.ElementTree import Element
from typing import Optional, Any, List
from .exceptions import OdxWarning

@dataclass(frozen=True)
class OdxDocFragment:
    doc_name : str
    doc_type : Optional[str]

    def __eq__(self, other) -> bool:
        if other is None:
            # if the other document fragment is not specified, we treat it as a wildcard...
            return True

        # the ODX spec says that the doctype can be ignored...
        return self.doc_name == other.doc_name

    def __hash__(self) -> int:
        # only the document name is relevant for the hash value
        return hash(self.doc_name) + hash(self.doc_type)

@dataclass(frozen=True)
class OdxLinkId:
    """The identifier of an ODX object.

    A full ODX ID comprises the document fragment where the object
    is located (i.e., the short name object beneath the root object of the
    currently read XMl file) plus a locally unique identifier for the object.

    OdxLinkIds are hashable.
    """

    #: An identifier that is (hopefully) locally unique within the
    #: relevant document fragment
    local_id : str

    #: The name and type of the document fragment to which the
    #: `local_id` is relative to
    doc_fragments : List[OdxDocFragment]

    def __hash__(self) -> int:
        # we do not hash about the document fragment here, because
        # document fragments are handled using separate sub-databases,
        # i.e. the same OdxId object can be put into all of them.
        return hash(self.local_id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, OdxLinkId):
            return False

        # we do take the document fragment into consideration here, because
        # document fragments are handled using separate sub-databases,
        # i.e. the same OdxId object can be put into all of them.
        return self.local_id == other.local_id

    @staticmethod
    def from_et(et: Element, doc_fragments: List[OdxDocFragment]) -> Optional["OdxLinkId"]:
        """Construct an OdxLinkId for a given XML node (ElementTree object).

        Returns None if the given XML node does not exhibit an ID.
        """

        local_id = et.attrib.get("ID")
        if local_id is None:
            return None

        return OdxLinkId(local_id, doc_fragments)

from typing import Dict, Union, Optional, Any
from dataclasses import dataclass, field, replace
from xml.etree.ElementTree import Element

@dataclass(frozen=True)
class OdxLinkRef:
    """A reference to an ODX object.

    OdxLinkRefs are similar to OdxLinkIds, with the difference that the document
    of the referred-to object does not need to be specified. (In this case, it
    means "use the document where the referring object is located in")
    """

    #: The local identifier of the object which is referred to
    ref_id : str

    #: The document fragments to which the `ref_id` refers to (in reverse order)
    ref_docs : List[OdxDocFragment]

    @staticmethod
    def from_et(et: Optional[Element], source_doc_frags: List[OdxDocFragment]) -> Optional["OdxLinkRef"]:
        """Construct an OdxLinkRef for a given XML node (ElementTree object).

        Returns None if the given XML node does not represent a reference.
        """

        if et is None:
            return None

        id_ref = et.attrib.get("ID-REF")
        if id_ref is None:
            return None

        doc_ref = et.attrib.get("DOCREF")
        doc_type = et.attrib.get("DOCTYPE")

        assert \
            (doc_ref is not None and doc_type is not None) or \
            (doc_ref is None and doc_type is None), \
            "DOCREF and DOCTYPE must both either be specified or omitted"

        # if the target document fragment is specified by the
        # reference, use it, else use the document fragment containing
        # the reference.
        if doc_ref is not None:
            doc_frags = [OdxDocFragment(doc_ref, doc_type)]
        else:
            doc_frags = source_doc_frags

        return OdxLinkRef(id_ref, doc_frags)

    @staticmethod
    def from_id(odxid: OdxLinkId) -> "OdxLinkRef":
        """Construct an OdxLinkRef for a given OdxLinkId.
        """
        return OdxLinkRef(odxid.local_id, odxid.doc_fragments)

    def __contains__(self, odx_id: OdxLinkId) -> bool:
        """
        Returns true iff a given OdxLinkId object is referenced.
        """

        # we must reference at to at least of the ID's document
        # fragments
        if not any([ref_doc in odx_id.doc_fragments for ref_doc in self.ref_docs]):
            return False

        # the local ID of the reference and the object ID must match
        return odx_id.local_id == self.ref_id

class OdxLinkDatabase:
    """
    A database holding all objects which ehibit OdxLinkIds

    This can resolve references to such.
    """

    def __init__(self) -> None:
        self._db: Dict[OdxDocFragment, Dict[OdxLinkId, Any]] = {}

    def resolve(self, ref: OdxLinkRef) -> Any:
        """
        Resolve a reference to an object

        If the database does not contain any object which is referred to, a
        KeyError exception is raised.
        """
        assert isinstance(ref, OdxLinkRef)

        odx_id = OdxLinkId(ref.ref_id, ref.ref_docs)
        for ref_frag in reversed(ref.ref_docs):
            doc_frag_db = self._db.get(ref_frag)
            if doc_frag_db is None:
                # No object featured by the database uses the document
                # fragment mentioned by the reference. This should not
                # happen for correct databases...
                warnings.warn(f"Warning: Unknown document fragment {ref_frag} "
                              f"when resolving reference {ref}", OdxWarning)
                continue

            obj = doc_frag_db.get(odx_id)
            if obj is not None:
                return obj

        raise KeyError(f"ODXLINK reference {ref} could not be resolved for any "
                       f"of the document fragments {ref.ref_docs}")

    def resolve_lenient(self, ref: OdxLinkRef) -> Optional[Any]:
        """
        Resolve a reference to an object

        If the database does not contain any object which is referred to, None
        is returned.
        """
        assert isinstance(ref, OdxLinkRef)

        odx_id = OdxLinkId(ref.ref_id, ref.ref_docs)
        for ref_frag in reversed(ref.ref_docs):
            doc_frag_db = self._db.get(ref_frag)
            if doc_frag_db is None:
                # No object featured by the database uses the document
                # fragment mentioned by the reference. This should not
                # happen for correct databases...
                warnings.warn(f"Warning: Unknown document fragment {ref_frag} "
                              f"when resolving reference {ref}", OdxWarning)
                continue

            obj = doc_frag_db.get(odx_id)
            if obj is not None:
                return obj

        return None

    def update(self, new_entries : Dict[OdxLinkId, Any]) -> None:
        """
        Add a bunch of new objects to the ODXLINK database.

        The argument needs to be an OdxLinkId -> object dictionary.
        """

        # put all new objects into the databases for the document
        # fragments which it specifies
        for odx_id, obj in new_entries.items():
            for doc_frag in odx_id.doc_fragments:
                if doc_frag not in self._db:
                    self._db[doc_frag] = dict()

                self._db[doc_frag][odx_id] = obj



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from collections import OrderedDict
from dataclasses import dataclass
from typing import Optional, Dict, Any, List

from .odxlink import OdxLinkRef, OdxLinkId, OdxLinkDatabase, OdxDocFragment
from .structures import BasicStructure
from .dataobjectproperty import DopBase, DataObjectProperty
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError
from .globals import logger


@dataclass
class MultiplexerCase:
    """This class represents a Case which represents multiple options in a Multiplexer."""

    short_name: str
    long_name: str
    structure_ref: OdxLinkRef
    lower_limit: str
    upper_limit: str

    def __post_init__(self) -> None:
        self._structure: Optional[BasicStructure] = None

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        self._structure = odxlinks.resolve(self.structure_ref)

    def __repr__(self) -> str:
        return (
            f"MultiplexerCase('{self.short_name}', "
            + ", ".join(
                [
                    f"lower_limit='{self.lower_limit}'",
                    f"upper_limit='{self.upper_limit}'",
                    f"structure_ref='{self.structure_ref}'",
                ]
            )
            + ")"
        )


@dataclass
class MultiplexerDefaultCase:
    """This class represents a Default Case, which is selected when there are no cases defined in the Multiplexer."""

    short_name: str
    long_name: str
    structure_ref: Optional[OdxLinkRef] = None

    def __post_init__(self) -> None:
        self._structure: Optional[BasicStructure] = None

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        if self.structure_ref is not None:
            self._structure = odxlinks.resolve(self.structure_ref)

    def __repr__(self) -> str:
        return (
            f"MultiplexerDefaultCase('{self.short_name}', "
            + ", ".join(
                [
                    f"structure_ref='{self.structure_ref}'",
                ]
            )
            + ")"
        )


@dataclass
class MultiplexerSwitchKey:
    """This class represents a Switch Key, which is used to select one of the cases defined in the Multiplexer."""

    byte_position: int
    bit_position: Optional[int]
    dop_ref: OdxLinkRef

    def __post_init__(self):
        self._dop: DataObjectProperty = None # type: ignore

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        dop = odxlinks.resolve(self.dop_ref)
        if isinstance(dop, DataObjectProperty):
            self._dop = dop
        else:
            logger.warning(
                f"DATA-OBJECT-PROP-REF '{self.dop_ref}' could not be resolved in SWITCH-KEY."
            )

    def __repr__(self) -> str:
        return (
            f"MultiplexerSwitchKey("
            + ", ".join(
                [
                    f"byte_position={self.byte_position}",
                    f"bit_position={self.bit_position}",
                    f"dop_ref={self.dop_ref}",
                ]
            )
            + ")"
        )


@dataclass
class Multiplexer(DopBase):
    """This class represents a Multiplexer (MUX) which are used to interpret data stream depending on the value
    of a switch-key (similar to switch-case statements in programming languages like C or Java)."""

    odx_id: OdxLinkId
    short_name: str
    long_name: str
    byte_position: int
    switch_key: MultiplexerSwitchKey
    default_case: Optional[MultiplexerDefaultCase] = None
    cases: Optional[List[MultiplexerCase]] = None

    @property
    def bit_length(self):
        return None

    def _get_case_limits(self, case: MultiplexerCase):
        key_type = self.switch_key._dop.physical_type.base_data_type
        lower_limit = key_type.make_from(case.lower_limit)
        upper_limit = key_type.make_from(case.upper_limit)
        return lower_limit, upper_limit

    def convert_physical_to_bytes(
            self,
            physical_value,
            encode_state: EncodeState,
            bit_position: int) -> bytes:

        if bit_position != 0:
            raise EncodeError(
                "Multiplexer must be aligned, i.e. bit_position=0, but "
                f"{self.short_name} was passed the bit position {bit_position}")

        if not isinstance(physical_value, dict) or len(physical_value) != 1:
            raise EncodeError("""Multiplexer should be defined as a dict
            with only one key equal to the desired case""")

        case_name, case_value = next(iter(physical_value.items()))
        key_pos = self.switch_key.byte_position
        case_pos = self.byte_position

        for case in (self.cases or []):
            if case.short_name == case_name:
                if case._structure:
                    case_bytes = case._structure.convert_physical_to_bytes(
                        case_value, encode_state, 0)
                else:
                    case_bytes = bytes()
                
                key_value, _ = self._get_case_limits(case)
                sk_bit_position = self.switch_key.bit_position
                sk_bit_position = sk_bit_position if sk_bit_position is not None else 0
                key_bytes = self.switch_key._dop.convert_physical_to_bytes(
                    key_value, encode_state, sk_bit_position)

                mux_len = max(
                    len(key_bytes) + key_pos,
                    len(case_bytes) + case_pos
                )
                mux_bytes = bytearray(mux_len)
                mux_bytes[key_pos:key_pos + len(key_bytes)] = key_bytes
                mux_bytes[case_pos:case_pos + len(case_bytes)] = case_bytes

                return bytes(mux_bytes)

        raise EncodeError(
            f"The case {case_name} is not found in Multiplexer {self.short_name}")

    def convert_bytes_to_physical(
            self,
            decode_state: DecodeState,
            bit_position: int = 0):

        if bit_position != 0:
            raise DecodeError(
                "Multiplexer must be aligned, i.e. bit_position=0, but "
                f"{self.short_name} was passed the bit position {bit_position}")
        byte_code = decode_state.coded_message[decode_state.next_byte_position:]
        key_decode_state = DecodeState(
            coded_message=byte_code[self.switch_key.byte_position:],
            parameter_value_pairs=[],
            next_byte_position=0
        )
        bit_position_int = self.switch_key.bit_position if self.switch_key.bit_position is not None else 0
        key_value, key_next_byte = \
            self.switch_key._dop.convert_bytes_to_physical(key_decode_state,
                                                           bit_position=bit_position_int)

        case_decode_state = DecodeState(
            coded_message=byte_code[self.byte_position:],
            parameter_value_pairs=[],
            next_byte_position=0
        )
        case_found = False
        case_next_byte = 0
        case_value = None
        for case in (self.cases or []):
            lower, upper = self._get_case_limits(case)
            if lower <= key_value and key_value <= upper:
                case_found = True
                if case._structure:
                    case_value, case_next_byte = case._structure.convert_bytes_to_physical(
                        case_decode_state)
                break

        if not case_found and self.default_case is not None:
            case_found = True
            if self.default_case._structure:
                case_value, case_next_byte = self.default_case._structure.convert_bytes_to_physical(
                    case_decode_state)

        if not case_found:
            raise DecodeError(
                f"Failed to find a matching case in {self.short_name} for value {key_value}")

        mux_value = OrderedDict({case.short_name: case_value})
        mux_next_byte = decode_state.next_byte_position + \
            max(
                key_next_byte + self.switch_key.byte_position,
                case_next_byte + self.byte_position
            )
        return mux_value, mux_next_byte

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {}
        odxlinks.update({self.odx_id: self})
        return odxlinks

    def _resolve_references(self, odxlinks: OdxLinkDatabase) -> None:
        self.switch_key._resolve_references(odxlinks)
        if self.default_case is not None:
            self.default_case._resolve_references(odxlinks)

        for case in (self.cases or []):
            case._resolve_references(odxlinks)

    def __repr__(self) -> str:
        return (
            f"Multiplexer('{self.short_name}', "
            + ", ".join(
                [
                    f"odx_id='{self.odx_id}'",
                    f"byte_position='{self.byte_position}'",
                    f"switch_key='{self.switch_key}'",
                    f"default_case='{self.default_case}'",
                    f"cases='{self.cases}'",
                ]
            )
            + ")"
        )


def read_switch_key_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    """Reads a Switch Key for a Multiplexer."""
    byte_position = int(et_element.findtext("BYTE-POSITION", "0"))
    assert byte_position is not None
    bit_position_str = et_element.findtext("BIT-POSITION")
    bit_position = int(bit_position_str) if bit_position_str is not None else None
    dop_ref = OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), doc_frags)
    assert dop_ref is not None

    return MultiplexerSwitchKey(
        byte_position=byte_position,
        bit_position=bit_position,
        dop_ref=dop_ref,
    )


def read_default_case_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    """Reads a Default Case for a Multiplexer."""
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")

    structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), doc_frags)

    return MultiplexerDefaultCase(
        short_name=short_name,
        long_name=long_name,
        structure_ref=structure_ref,
    )


def read_case_from_odx(et_element, doc_frags: List[OdxDocFragment]):
    """Reads a Case for a Multiplexer."""
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), doc_frags)
    assert structure_ref is not None
    lower_limit = et_element.findtext("LOWER-LIMIT")
    upper_limit = et_element.findtext("UPPER-LIMIT")

    return MultiplexerCase(
        short_name=short_name,
        long_name=long_name,
        structure_ref=structure_ref,
        lower_limit=lower_limit,
        upper_limit=upper_limit,
    )


def read_mux_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
    -> Multiplexer:
    """Reads a Multiplexer from Diag Layer."""
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    byte_position = int(et_element.findtext("BYTE-POSITION", "0"))
    switch_key = read_switch_key_from_odx(et_element.find("SWITCH-KEY"),
                                          doc_frags)

    default_case = None
    if et_element.find("DEFAULT-CASE") is not None:
        default_case = read_default_case_from_odx(
            et_element.find("DEFAULT-CASE"), doc_frags)

    cases = []
    if et_element.find("CASES") is not None:
        cases = [
            read_case_from_odx(el, doc_frags) for el in et_element.find("CASES").iterfind("CASE")
        ]

    logger.debug("Parsing MUX " + short_name)

    return Multiplexer(
        odx_id=odx_id,
        short_name=short_name,
        long_name=long_name,
        byte_position=byte_position,
        switch_key=switch_key,
        default_case=default_case,
        cases=cases,
    )



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH


from typing import List, Tuple, Union

from ..exceptions import EncodeError, DecodeError
from ..globals import logger
from ..odxtypes import DataType

from .compumethodbase import CompuMethod
from .limit import IntervalType, Limit


class TabIntpCompuMethod(CompuMethod):
    """
    A compu method of type Tab Interpolated is used for linear interpolation.

    A `TabIntpCompuMethod` is defined by a set of points. Each point is an (internal, physical) value pair.
    When converting from internal to physical or vice-versa, the result is linearly interpolated.

    The function defined by a `TabIntpCompuMethod` is similar to the one of a `ScaleLinearCompuMethod` with the following differences:

    * `TabIntpCompuMethod`s are always continuous whereas `ScaleLinearCompuMethod` might have jumps
    * `TabIntpCompuMethod`s are always invertible: Even if the linear interpolation is not monotonic, the first matching interval is taken.

    Refer to ASAM MCD-2 D (ODX) Specification, section 7.3.6.6.8 for details.

    Examples
    --------

    Create a TabIntpCompuMethod defined by the points (0, -1), (10, 1), (30, 2)::

        method = TabIntpCompuMethod(
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32,
            internal_points=[0, 10, 30],
            physical_points=[-1, 1, 2]
        )

    Note that the points are given as two lists. The equivalent odx definition is::

        <COMPU-METHOD>
            <CATEGORY>TAB-INTP</CATEGORY>
            <COMPU-INTERNAL-TO-PHYS>
                <COMPU-SCALES>
                    <COMPU-SCALE>
                        <LOWER-LIMIT INTERVAL-TYPE = "CLOSED">0</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>-1</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                    <COMPU-SCALE>
                        <LOWER-LIMIT INTERVAL-TYPE = "CLOSED">10</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>1</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                    <COMPU-SCALE>
                        <LOWER-LIMIT INTERVAL-TYPE = "CLOSED">30</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>2</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                </COMPU-SCALES>
            </COMPU-INTERNAL-TO-PHYS>
        </COMPU-METHOD>

    """

    def __init__(self,
                 internal_type: Union[DataType, str],
                 physical_type: Union[DataType, str],
                 internal_points: List[Union[float, int]],
                 physical_points: List[Union[float, int]]):
        super().__init__(internal_type, physical_type, "TAB-INTP")

        self.internal_points = internal_points
        self.physical_points = physical_points

        self._physical_lower_limit = Limit(
            min(physical_points), IntervalType.CLOSED)
        self._physical_upper_limit = Limit(
            max(physical_points), IntervalType.CLOSED)

        logger.debug("Created compu method of type tab interpolated !")
        self._assert_validity()

    @property
    def physical_lower_limit(self) -> Limit:
        return self._physical_lower_limit

    @property
    def physical_upper_limit(self) -> Limit:
        return self._physical_upper_limit

    def _assert_validity(self) -> None:
        assert len(self.internal_points) == len(self.physical_points)

        assert self.internal_type in [DataType.A_INT32, DataType.A_UINT32,
                                      DataType.A_FLOAT32, DataType.A_FLOAT64], \
            ("Internal data type of tab-intp compumethod must be one of"
             " [DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64]")
        assert self.physical_type in [DataType.A_INT32, DataType.A_UINT32,
                                      DataType.A_FLOAT32, DataType.A_FLOAT64], \
            ("Physical data type of tab-intp compumethod must be one of"
             " [DataType.A_INT32, DataType.A_UINT32, DataType.A_FLOAT32, DataType.A_FLOAT64]")

    def _piecewise_linear_interpolate(self,
                                      x: Union[int, float],
                                      points: List[Tuple[Union[int, float], Union[int, float]]]) \
            -> Union[float, None]:
        for ((x0, y0), (x1, y1)) in zip(points[:-1], points[1:]):
            if x0 <= x and x <= x1:
                return y0 + (x - x0) * (y1 - y0) / (x1 - x0)

        return None

    def convert_physical_to_internal(self, physical_value: Union[int, float]) -> Union[int, float]:
        reference_points = list(zip(
            self.physical_points, self.internal_points))
        result = self._piecewise_linear_interpolate(
            physical_value, reference_points)

        if result is None:
            raise EncodeError(f"Internal value {physical_value} must be inside the range"
                              f" [{min(self.physical_points)}, {max(self.physical_points)}]")
        res = self.internal_type.make_from(result)
        assert isinstance(res, (int, float))
        return res

    def convert_internal_to_physical(self, internal_value: Union[int, float]) -> Union[int, float]:
        reference_points = list(zip(
            self.internal_points, self.physical_points))
        result = self._piecewise_linear_interpolate(
            internal_value, reference_points)

        if result is None:
            raise DecodeError(f"Internal value {internal_value} must be inside the range"
                              f" [{min(self.internal_points)}, {max(self.internal_points)}]")
        res = self.physical_type.make_from(result)
        assert isinstance(res, (int, float))
        return res

    def is_valid_physical_value(self, physical_value: Union[int, float]) -> bool:
        return min(self.physical_points) <= physical_value and physical_value <= max(self.physical_points)

    def is_valid_internal_value(self, internal_value: Union[int, float]) -> bool:
        return min(self.internal_points) <= internal_value and internal_value <= max(self.internal_points)



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from pathlib import Path
from typing import List, Optional
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
from itertools import chain
from zipfile import ZipFile

from .utils import short_name_as_id
from .odxlink import OdxLinkDatabase
from .diaglayer import DiagLayer, DiagLayerContainer, read_diag_layer_container_from_odx
from .comparam_subset import ComparamSubset, read_comparam_subset_from_odx
from .globals import logger
from .nameditemlist import NamedItemList

def version(v: str):
    return tuple(map(int, (v.split("."))))

class Database:
    """This class internalizes the diagnostic database for various ECUs
    described by a collection of ODX files which are usually collated
    into a single PDX file.
    """

    def __init__(self,
                 pdx_zip: Optional[ZipFile] = None,
                 odx_d_file_name: Optional[str] = None):

        if pdx_zip is None and odx_d_file_name is None:
            # create an empty database object
            self._diag_layer_containers = NamedItemList(short_name_as_id, [])
            self._comparam_subsets = NamedItemList(short_name_as_id, [])
            return

        if pdx_zip is not None and odx_d_file_name is not None:
            raise TypeError("The 'pdx_zip' and 'odx_d_file_name' parameters are mutually exclusive")

        documents: List[Element] = []
        if pdx_zip is not None:
            names = list(pdx_zip.namelist())
            names.sort()
            for zip_member in names:
                # file name can end with .odx, .odx-d, .odx-c, .odx-cs, .odx-e, .odx-f, .odx-fd, .odx-m, .odx-v
                # We could test for all that, or just make sure suffix starts with .odx
                if Path(zip_member).suffix.startswith(".odx"):
                    logger.info(f"Processing the file {zip_member}")
                    d = pdx_zip.read(zip_member)
                    root = ElementTree.fromstring(d)
                    documents.append(root)

        elif odx_d_file_name is not None:
            documents.append(ElementTree.parse(odx_d_file_name).getroot())

        dlcs: List[DiagLayerContainer] = []
        comparam_subsets: List[ComparamSubset] = []
        for root in documents:
            # ODX spec version
            model_version = version(root.attrib.get('MODEL-VERSION', '2.0'))
            dlc = root.find("DIAG-LAYER-CONTAINER")
            if dlc is not None:
                dlcs.append(read_diag_layer_container_from_odx(
                    dlc
                ))
            # In ODX 2.0 there was only COMPARAM-SPEC
            # In ODX 2.2 content of COMPARAM-SPEC was renamed to COMPARAM-SUBSET
            # and COMPARAM-SPEC becomes a container for PROT-STACKS
            # and a PROT-STACK references a list of COMPARAM-SUBSET
            if model_version >= version('2.2'):
                subset = root.find("COMPARAM-SUBSET")
                if subset is not None:
                    comparam_subsets.append(read_comparam_subset_from_odx(subset))
            else:
                subset = root.find("COMPARAM-SPEC")
                if subset is not None:
                    comparam_subsets.append(read_comparam_subset_from_odx(subset))

        self._diag_layer_containers = NamedItemList(short_name_as_id, dlcs)
        self._diag_layer_containers.sort(key=short_name_as_id)
        self._comparam_subsets = NamedItemList(short_name_as_id, comparam_subsets)
        self._comparam_subsets.sort(key=short_name_as_id)
        self.finalize_init()

    def finalize_init(self) -> None:
        # Create wrapper objects
        self._diag_layers = NamedItemList(
            short_name_as_id,
            chain(*[dlc.diag_layers for dlc in self.diag_layer_containers]))
        self._ecus = NamedItemList(
            short_name_as_id,
            chain(*[dlc.ecu_variants for dlc in self.diag_layer_containers]))

        # Build odxlinks
        self._odxlinks = OdxLinkDatabase()

        for subset in self.comparam_subsets:
            self.odxlinks.update(subset._build_odxlinks())

        for dlc in self.diag_layer_containers:
            self.odxlinks.update(dlc._build_odxlinks())

        for dl in self.diag_layers:
            self.odxlinks.update(dl._build_odxlinks())

        # Resolve references
        for subset in self.comparam_subsets:
            subset._resolve_references(self.odxlinks)
        for dlc in self.diag_layer_containers:
            dlc._resolve_references(self.odxlinks)

        for dl_type_name in ["ECU-SHARED-DATA", "PROTOCOL", "FUNCTIONAL-GROUP", "BASE-VARIANT", "ECU-VARIANT"]:
            for dl in self.diag_layers:
                if dl.variant_type == dl_type_name:
                    dl._resolve_references(self.odxlinks)

    @property
    def odxlinks(self) -> OdxLinkDatabase:
        """A map from odx_id to object"""
        return self._odxlinks

    @property
    def ecus(self) -> NamedItemList[DiagLayer]:
        """ECU-variants defined in the data base"""
        return self._ecus

    @property
    def diag_layers(self) -> NamedItemList[DiagLayer]:
        """all diagnostic layers defined in the data base"""
        return self._diag_layers

    @property
    def diag_layer_containers(self):
        return self._diag_layer_containers

    @diag_layer_containers.setter
    def diag_layer_containers(self, value):
        self._diag_layer_containers = value

    @property
    def comparam_subsets(self):
        return self._comparam_subsets


# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from typing import Any, Dict, List

from .parameters import read_parameter_from_odx
from .utils import read_description_from_odx
from .odxlink import OdxLinkId, OdxDocFragment
from .structures import BasicStructure
from .globals import logger

@dataclass
class EnvironmentData(BasicStructure):
    """This class represents Environment Data that describes the circumstances in which the error occurred."""

    def __init__(
        self,
        odx_id,
        short_name,
        parameters,
        long_name=None,
        description=None,
    ):
        super().__init__(
            odx_id, short_name, parameters, long_name=long_name, description=description
        )

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {self.odx_id: self}
        return odxlinks

    def __repr__(self) -> str:
        return (
            f"EnvironmentData('{self.short_name}', "
            + ", ".join([f"odx_id='{self.odx_id}'", f"parameters='{self.parameters}'"])
            + ")"
        )


def read_env_data_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
    -> EnvironmentData:
    """Reads Environment Data from Diag Layer."""
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    description = read_description_from_odx(et_element.find("DESC"))
    parameters = [
        read_parameter_from_odx(et_parameter, doc_frags)
        for et_parameter in et_element.iterfind("PARAMS/PARAM")
    ]
    logger.debug("Parsing ENV-DATA " + short_name)

    env_data = EnvironmentData(
        odx_id,
        short_name,
        parameters=parameters,
        long_name=long_name,
        description=description,
    )

    return env_data



# SPDX-License-Identifier: MIT
# Copyright (c) 2022 MBition GmbH

from dataclasses import dataclass
from typing import List, Dict, Optional, Any

from .odxlink import OdxLinkRef, OdxLinkId, OdxDocFragment
from .globals import logger
from .dataobjectproperty import DopBase

from .envdata import EnvironmentData, read_env_data_from_odx
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError

class EnvironmentDataDescription(DopBase):
    """This class represents Environment Data Description, which is a complex DOP
    that is used to define the interpretation of environment data."""

    def __init__(self,
                 env_datas: List[EnvironmentData],
                 env_data_refs: List[OdxLinkRef],
                 param_snref: Optional[str] = None,
                 param_snpathref: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)

        self.bit_length = None
        self.env_datas = env_datas
        self.env_data_refs = env_data_refs
        self.param_snref = param_snref
        self.param_snpathref = param_snpathref

        # TODO: resolve the references!

    def _build_odxlinks(self) -> Dict[OdxLinkId, Any]:
        odxlinks = {}
        odxlinks[self.odx_id] = self

        for ed in self.env_datas:
            odxlinks.update(ed._build_odxlinks())

        return odxlinks

    def __repr__(self) -> str:
        return (
            f"EnvironmentDataDescription('{self.short_name}', "
            + ", ".join(
                [
                    f"odx_id='{self.odx_id}'",
                    f"param_snref='{self.param_snref}'",
                    f"param_snpathref='{self.param_snpathref}'",
                    f"env_data_refs='{self.env_data_refs}'",
                ]
            )
            + ")"
        )

    def convert_physical_to_bytes(self, physical_value, encode_state: EncodeState, bit_position: int) -> bytes:
        """Convert the physical value into bytes.

        Since environmental data is supposed to never appear on the
        wire, this method just raises an EncodeError exception.
        """
        raise EncodeError("EnvironmentDataDescription DOPs cannot be encoded or decoded")

    def convert_bytes_to_physical(self, decode_state: DecodeState, bit_position: int = 0):
        """Extract the bytes from the PDU and convert them to the physical value.

        Since environmental data is supposed to never appear on the
        wire, this method just raises an DecodeError exception.
        """
        raise DecodeError("EnvironmentDataDescription DOPs cannot be encoded or decoded")


def read_env_data_desc_from_odx(et_element, doc_frags: List[OdxDocFragment]) \
    -> EnvironmentDataDescription:
    """Reads Environment Data Description from Diag Layer."""
    odx_id = OdxLinkId.from_et(et_element, doc_frags)
    assert odx_id is not None
    short_name = et_element.findtext("SHORT-NAME")
    long_name = et_element.findtext("LONG-NAME")
    param_snref = None
    if et_element.find("PARAM-SNREF") is not None:
        param_snref = et_element.find("PARAM-SNREF").get("SHORT-NAME")
    param_snpathref = None
    if et_element.find("PARAM-SNPATHREF") is not None:
        param_snpathref = et_element.find("PARAM-SNPATHREF").get("SHORT-NAME-PATH")
    env_data_refs = []
    for env_data_ref in et_element.iterfind("ENV-DATA-REFS/ENV-DATA-REF"):
        ref = OdxLinkRef.from_et(env_data_ref, doc_frags)
        assert ref is not None
        env_data_refs.append(ref)

    # ODX 2.0.0 says ENV-DATA-DESC could contain a list of ENV-DATAS
    env_datas = []
    for env_data_elem in et_element.iterfind("ENV-DATAS/ENV-DATA"):
        env_data = read_env_data_from_odx(env_data_elem, doc_frags)
        env_datas.append(env_data)

    logger.debug("Parsing ENV-DATA-DESC " + short_name)

    return EnvironmentDataDescription(
        odx_id=odx_id,
        short_name=short_name,
        long_name=long_name,
        param_snref=param_snref,
        param_snpathref=param_snpathref,
        env_datas=env_datas,
        env_data_refs=env_data_refs,
    )
